import unittest
from os import urandom

from nostr.key import PrivateKey, PublicKey


class TestPrivateKey(unittest.TestCase):
    def test_eq_true(self):
        """__eq__ should return True when PrivateKeys are equal."""
        pk1 = PrivateKey()
        pk2 = PrivateKey(pk1.raw_secret)
        self.assertEqual(pk1, pk2)

    def test_eq_false(self):
        """__eq__ should return False when PrivateKeys are not equal."""
        pk1 = PrivateKey()
        pk2 = PrivateKey()
        self.assertNotEqual(pk1.raw_secret, pk2.raw_secret)
        self.assertNotEqual(pk1, pk2)

    def test_from_nsec(self):
        """PrivateKey.from_nsec should yield the source's raw_secret."""
        pk1 = PrivateKey()
        pk2 = PrivateKey.from_nsec(pk1.bech32())
        self.assertEqual(pk1.raw_secret, pk2.raw_secret)

    def test_from_hex2(self):
        pk1 = PrivateKey.from_hex(
            "964b29795d621cdacf05fd94fb23206c88742db1fa50b34d7545f3a2221d8124"
        )
        pub1 = PublicKey(pk1)
        self.assertEqual(
            pub1.hex(),
            "da15317263858ad496a21c79c6dc5f5cf9af880adf3a6794dbbf2883186c9d81",
        )

        pub1 = PublicKey(
            "da15317263858ad496a21c79c6dc5f5cf9af880adf3a6794dbbf2883186c9d81"
        )
        self.assertEqual(
            pub1.hex(),
            "da15317263858ad496a21c79c6dc5f5cf9af880adf3a6794dbbf2883186c9d81",
        )

        pub1 = PublicKey(
            "da15317263858ad496a21c79c6dc5f5cf9af880adf3a6794dbbf2883186c9d81"
        )
        self.assertEqual(
            pub1.hex(),
            "da15317263858ad496a21c79c6dc5f5cf9af880adf3a6794dbbf2883186c9d81",
        )
        self.assertEqual(
            pk1.public_key.hex(),
            "da15317263858ad496a21c79c6dc5f5cf9af880adf3a6794dbbf2883186c9d81",
        )
        self.assertEqual(
            pk1.bech32(),
            "nsec1je9jj72avgwd4nc9lk20kgeqdjy8gtd3lfgtxnt4ghe6ygsasyjq7kh6c4",
        )
        self.assertEqual(
            pub1.bech32(),
            "npub1mg2nzunrsk9df94zr3uudhzltnu6lzq2muax09xmhu5gxxrvnkqsvpjg3p",
        )

    def test_from_nsec2(self):
        pk1 = PrivateKey.from_nsec(
            "nsec1je9jj72avgwd4nc9lk20kgeqdjy8gtd3lfgtxnt4ghe6ygsasyjq7kh6c4"
        )
        pub1 = PublicKey(pk1)
        self.assertEqual(
            pub1.hex(),
            "da15317263858ad496a21c79c6dc5f5cf9af880adf3a6794dbbf2883186c9d81",
        )
        self.assertEqual(
            pk1.bech32(),
            "nsec1je9jj72avgwd4nc9lk20kgeqdjy8gtd3lfgtxnt4ghe6ygsasyjq7kh6c4",
        )
        self.assertEqual(
            pub1.bech32(),
            "npub1mg2nzunrsk9df94zr3uudhzltnu6lzq2muax09xmhu5gxxrvnkqsvpjg3p",
        )

    def test_pubkey_from_nsec(self):
        pub1 = PublicKey.from_npub(
            "npub1mg2nzunrsk9df94zr3uudhzltnu6lzq2muax09xmhu5gxxrvnkqsvpjg3p"
        )
        self.assertEqual(
            pub1.hex(),
            "da15317263858ad496a21c79c6dc5f5cf9af880adf3a6794dbbf2883186c9d81",
        )
        self.assertEqual(
            pub1.bech32(),
            "npub1mg2nzunrsk9df94zr3uudhzltnu6lzq2muax09xmhu5gxxrvnkqsvpjg3p",
        )

    def test_schnorr_signature(self):
        private_key = PrivateKey()
        message = urandom(32)
        sig = private_key.sign(message, urandom(32))
        self.assertTrue(private_key.public_key.verify(sig, message))
        sig = private_key.sign(message)
        self.assertTrue(private_key.public_key.verify(sig, message))

    def test_shared_secret(self):
        sender_pk = "29307c4354b7d9d311d2cec4878c0de56c93a921d300273c19577e9004de3c9f"
        recipient_pk = (
            "4138d1b6dde34f81c38cef2630429e85847dd5b70508e37f53c844f66f19f983"
        )
        recipient_pub = (
            "a1db8e8b047e1350958a55e0a853151d0e1f685fa5cf3772e01bccc5aa5cb2eb"
        )
        sender_pub = "f3c25355c29f64ea8e9b4e11b583ac0a7d0d8235f156cffec2b73e5756aab206"

        sender_pubkey = PrivateKey.from_hex(sender_pk).public_key.hex()
        recipient_pubkey = PrivateKey.from_hex(recipient_pk).public_key.hex()
        self.assertEqual(sender_pubkey, sender_pub)
        self.assertEqual(recipient_pubkey, recipient_pub)

        shared_secret1 = PrivateKey.from_hex(sender_pk).compute_shared_secret(
            recipient_pub
        )
        shared_secret2 = PrivateKey.from_hex(recipient_pk).compute_shared_secret(
            sender_pub
        )
        self.assertEqual(
            shared_secret1.hex(),
            "646570d4716e0c7e4106788f113a410d5b647225dca3b47ef98bedb64c8044e1",
        )
        self.assertEqual(
            shared_secret2.hex(),
            "646570d4716e0c7e4106788f113a410d5b647225dca3b47ef98bedb64c8044e1",
        )



import unittest

from nostr.event import Event, EventKind
from nostr.filter import Filter, Filters
from nostr.key import PrivateKey


class TestFilter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.pk1 = PrivateKey()
        cls.pk2 = PrivateKey()

        """ pk1 kicks off a thread and interacts with pk2 """
        cls.pk1_thread = [
            # Note posted by pk1
            Event(public_key=cls.pk1.public_key.hex(), content="pk1's first note!"),
        ]
        cls.pk1_thread.append(
            # Note posted by pk2 in response to pk1's note
            Event(
                public_key=cls.pk2.public_key.hex(),
                content="Nice to see you here, pk1!",
                tags=[
                    [
                        "e",
                        cls.pk1_thread[0].id,
                    ],  # Replies reference which note they're directly responding to
                    [
                        "p",
                        cls.pk1.public_key.hex(),
                    ],  # Replies reference who they're responding to
                ],
            )
        )
        cls.pk1_thread.append(
            # Next response note by pk1 continuing thread with pk2
            Event(
                public_key=cls.pk1.public_key.hex(),
                content="Thanks! Glad you're here, too, pk2!",
                tags=[
                    ["e", cls.pk1_thread[0].id],  # Threads reference the original note
                    [
                        "e",
                        cls.pk1_thread[-1].id,
                    ],  # Replies reference which note they're directly responding to
                    [
                        "p",
                        cls.pk2.public_key.hex(),
                    ],  # Replies reference who they're responding to
                ],
            )
        )

        """ pk2 starts a new thread but no one responds """
        cls.pk2_thread = [
            # Note posted by pk2
            Event(public_key=cls.pk2.public_key.hex(), content="pk2's first note!")
        ]
        cls.pk2_thread.append(
            # pk2's self-reply
            Event(
                public_key=cls.pk2.public_key.hex(),
                content="So... I guess no one's following me.",
                tags=[["e", cls.pk2_thread[0].id]],
            )
        )

        """ pk1 DMs pk2 """
        cls.pk1_pk2_dms = [
            # DM sent by pk1 to pk2
            Event(
                public_key=cls.pk1.public_key.hex(),
                content="Hey pk2, here's a secret",
                tags=[["p", cls.pk2.public_key.hex()]],
                kind=EventKind.ENCRYPTED_DIRECT_MESSAGE,
            ),
            Event(
                public_key=cls.pk2.public_key.hex(),
                content="Thanks! I'll keep it secure.",
                tags=[["p", cls.pk1.public_key.hex()]],
                kind=EventKind.ENCRYPTED_DIRECT_MESSAGE,
            ),
        ]

    def test_match_by_event_id(self):
        """Should match Events by event_id."""
        filter = Filter(
            event_ids=[self.pk1_thread[0].id],
        )
        assert filter.matches(self.pk1_thread[0])

        # None of the others should match
        for event in self.pk1_thread[1:] + self.pk2_thread + self.pk1_pk2_dms[1:]:
            assert filter.matches(event) is False

    def test_multiple_values_in_same_tag(self):
        """Should treat multiple tag values as OR searches."""
        filter = Filter(
            event_ids=[
                self.pk1_thread[0].id,
                self.pk1_pk2_dms[0].id,
                "some_other_event_id",
            ],
        )
        assert filter.matches(self.pk1_thread[0])
        assert filter.matches(self.pk1_pk2_dms[0])

        # None of the others should match
        for event in self.pk1_thread[1:] + self.pk2_thread + self.pk1_pk2_dms[1:]:
            assert filter.matches(event) is False

    def test_match_by_kinds(self):
        """Should match Events by kind."""
        filter = Filter(
            kinds=[EventKind.TEXT_NOTE],
        )

        # Both threads should match
        for event in self.pk1_thread + self.pk2_thread:
            assert filter.matches(event)

        # DMs should not match
        for event in self.pk1_pk2_dms:
            assert filter.matches(event) is False

        # Now allow either kind
        filter = Filter(
            kinds=[EventKind.TEXT_NOTE, EventKind.ENCRYPTED_DIRECT_MESSAGE],
        )

        # Now everything should match
        for event in self.pk1_thread + self.pk2_thread + self.pk1_pk2_dms:
            assert filter.matches(event)

    def test_match_by_authors(self):
        """Should match Events by author."""
        filter = Filter(authors=[self.pk1.public_key.hex()])

        # Everything sent by pk1 should match
        for event in [
            event
            for event in (self.pk1_thread + self.pk2_thread + self.pk1_pk2_dms)
            if event.public_key == self.pk1.public_key.hex()
        ]:
            assert filter.matches(event)

        # None of pk2's should match
        for event in [
            event
            for event in (self.pk1_thread + self.pk2_thread + self.pk1_pk2_dms)
            if event.public_key == self.pk2.public_key.hex()
        ]:
            assert filter.matches(event) is False

    def test_match_by_event_refs(self):
        """Should match Events by event_ref 'e' tags."""
        filter = Filter(
            event_refs=[self.pk1_thread[0].id],
        )

        # All replies to pk1's initial note should match (even pk1's reply at the end)
        assert filter.matches(self.pk1_thread[1])
        assert filter.matches(self.pk1_thread[2])

        # Everything else should not match
        for event in [self.pk1_thread[0]] + self.pk2_thread + self.pk1_pk2_dms:
            assert filter.matches(event) is False

    def test_match_by_pubkey_refs(self):
        """Should match Events by pubkey_ref 'p' tags."""
        filter = Filter(
            pubkey_refs=[self.pk1_thread[0].public_key],
        )

        # pk2's reply in pk1's thread should match
        assert filter.matches(self.pk1_thread[1])

        # pk2's DM reply to pk1 should match
        assert filter.matches(self.pk1_pk2_dms[1])

        # Everything else should not match
        for event in (
            [self.pk1_thread[0], self.pk1_thread[2]]
            + self.pk2_thread
            + [self.pk1_pk2_dms[0]]
        ):
            assert filter.matches(event) is False

    def test_match_by_arbitrary_single_letter_tag(self):
        """Should match NIP-12 arbitrary single-letter tags."""
        filter = Filter()
        filter.add_arbitrary_tag("x", ["oranges"])

        # None of our Events match
        for event in self.pk1_thread + self.pk2_thread + self.pk1_pk2_dms:
            assert filter.matches(event) is False

        # A new Event that has the target tag but the wrong value should not match
        event = Event(
            public_key=self.pk1.public_key.hex(),
            content="Additional event to test with",
            tags=[["x", "bananas"]],
        )
        assert filter.matches(event) is False

        # But a new Event that includes the target should match
        event = Event(
            public_key=self.pk1.public_key.hex(),
            content="Additional event to test with",
            tags=[["x", "oranges"]],
        )
        assert filter.matches(event)

        # Filter shouldn't care if there are other extraneous values
        event.tags.append(["x", "pizza"])
        assert filter.matches(event)

        event.tags.append(["y", "honey badger"])
        assert filter.matches(event)

    def test_match_by_arbitrary_multi_letter_tag(self):
        """Should match any arbitrary multi-letter tag."""
        filter = Filter()
        filter.add_arbitrary_tag("favorites", ["bitcoin"])

        # None of our Events match
        for event in self.pk1_thread + self.pk2_thread + self.pk1_pk2_dms:
            assert filter.matches(event) is False

        # A new Event that has the target tag but the wrong value should not match
        event = Event(
            public_key=self.pk1.public_key.hex(),
            content="Additional event to test with",
            tags=[["favorites", "shitcoin"]],
        )
        assert filter.matches(event) is False

        # But a new Event that includes the target should match
        event = Event(
            public_key=self.pk1.public_key.hex(),
            content="Additional event to test with",
            tags=[["favorites", "bitcoin"]],
        )
        assert filter.matches(event)

        # Filter shouldn't care if there are other extraneous values
        event.tags.append(["favorites", "sats"])
        assert filter.matches(event)

        event.tags.append(["foo", "bar"])
        assert filter.matches(event)

    def test_match_by_delegation_tag(self):
        """should match on delegation tag.

        Note: this is to demonstrate that it works w/out special handling, but
        arguably Delegation filtering should have its own explicit Filter support.
        """
        filter = Filter()

        # Search just for the delegator's pubkey
        # (only aspect of delegation search that is supported this way)
        filter.add_arbitrary_tag(
            "delegation",
            ["8e0d3d3eb2881ec137a11debe736a9086715a8c8beeeda615780064d68bc25dd"],
        )

        # None of our Events match
        for event in self.pk1_thread + self.pk2_thread + self.pk1_pk2_dms:
            assert filter.matches(event) is False

        # A new Event that has the target tag but the wrong value should not match
        event = Event(
            public_key=self.pk1.public_key.hex(),
            content="Additional event to test with",
            tags=[
                [
                    "delegation",
                    "some_other_delegators_pubkey",
                    "kind=1&created_at<1675721813",
                    "cbc49c65fe04a3181d72fb5a9f1c627e329d5f45d300a2dfed1c3e788b7834dad48a6d27d8e244af39c77381334ede97d4fd15abe80f35fda695fd9bd732aa1e",  # noqa: E501
                ]
            ],
        )
        assert filter.matches(event) is False

        # But a new Event that includes the target should match
        event = Event(
            public_key=self.pk1.public_key.hex(),
            content="Additional event to test with",
            tags=[
                [
                    "delegation",
                    "8e0d3d3eb2881ec137a11debe736a9086715a8c8beeeda615780064d68bc25dd",
                    "kind=1&created_at<1675721813",
                    "cbc49c65fe04a3181d72fb5a9f1c627e329d5f45d300a2dfed1c3e788b7834dad48a6d27d8e244af39c77381334ede97d4fd15abe80f35fda695fd9bd732aa1e",  # noqa: E501
                ]
            ],
        )
        assert filter.matches(event)

        # Filter shouldn't care if there are other extraneous values
        event.tags.append(["favorites", "sats"])
        assert filter.matches(event)

        event.tags.append(["foo", "bar"])
        assert filter.matches(event)

    def test_match_by_authors_and_kinds(self):
        """Should match Events by authors AND kinds."""
        filter = Filter(
            authors=[self.pk1.public_key.hex()],
            kinds=[EventKind.TEXT_NOTE],
        )

        # Should match pk1's notes but not pk2's reply
        assert filter.matches(self.pk1_thread[0])
        assert filter.matches(self.pk1_thread[1]) is False
        assert filter.matches(self.pk1_thread[2])

        # Should not match anything else
        for event in self.pk2_thread + self.pk1_pk2_dms:
            assert filter.matches(event) is False

        # Typical search to get all Events sent by a pubkey
        filter = Filter(
            authors=[self.pk1.public_key.hex()],
            kinds=[EventKind.TEXT_NOTE, EventKind.ENCRYPTED_DIRECT_MESSAGE],
        )

        # Should still match pk1's notes but not pk2's reply
        assert filter.matches(self.pk1_thread[0])
        assert filter.matches(self.pk1_thread[1]) is False
        assert filter.matches(self.pk1_thread[2])

        # Should not match any of pk2's solo thread
        assert filter.matches(self.pk2_thread[0]) is False
        assert filter.matches(self.pk2_thread[1]) is False

        # Should match pk1's DM but not pk2's DM reply
        assert filter.matches(self.pk1_pk2_dms[0])
        assert filter.matches(self.pk1_pk2_dms[1]) is False

    def test_match_by_kinds_and_pubkey_refs(self):
        """Should match Events by kind AND pubkey_ref 'p' tags."""
        filter = Filter(
            kinds=[EventKind.TEXT_NOTE],
            pubkey_refs=[self.pk2.public_key.hex()],
        )

        # Only pk1's reply to pk2 should match
        assert filter.matches(self.pk1_thread[2])

        # Should not match anything else
        for event in self.pk1_thread[:1] + self.pk2_thread + self.pk1_pk2_dms:
            assert filter.matches(event) is False

        # Typical search to get all Events sent to a pubkey
        filter = Filter(
            kinds=[EventKind.TEXT_NOTE, EventKind.ENCRYPTED_DIRECT_MESSAGE],
            pubkey_refs=[self.pk2.public_key.hex()],
        )

        # pk1's reply to pk2 should match
        assert filter.matches(self.pk1_thread[2])

        # pk2's DM to pk1 should match
        assert filter.matches(self.pk1_pk2_dms[0])

        # Should not match anything else
        for event in self.pk1_thread[:1] + self.pk2_thread + self.pk1_pk2_dms[1:]:
            assert filter.matches(event) is False

    def test_event_refs_json(self):
        """Should insert event_refs as "#e" in json."""
        filter = Filter(event_refs=["some_event_id"])
        assert "#e" in filter.to_json_object().keys()
        assert "e" not in filter.to_json_object().keys()

    def test_pubkey_refs_json(self):
        """Should insert pubkey_refs as "#p" in json."""
        filter = Filter(pubkey_refs=["some_pubkey"])
        assert "#p" in filter.to_json_object().keys()
        assert "p" not in filter.to_json_object().keys()

    def test_arbitrary_single_letter_json(self):
        """Should prefix NIP-12 arbitrary single-letter tags with "#" in json."""
        filter = Filter()
        filter.add_arbitrary_tag("x", ["oranges"])
        assert "#x" in filter.to_json_object().keys()
        assert "x" not in filter.to_json_object().keys()

    def test_arbitrary_multi_letter_json(self):
        """Should include arbitrary multi-letter tags as-is in json."""
        filter = Filter()
        filter.add_arbitrary_tag("foo", ["bar"])
        assert "foo" in filter.to_json_object().keys()


# Inherit from TestFilter to get all the same test data
class TestFilters(TestFilter):
    def test_match_by_authors_or_pubkey_refs(self):
        """Should match on authors or pubkey_refs."""
        # Typical filters for anything sent by or to a pubkey
        filter1 = Filter(
            authors=[self.pk1.public_key.hex()],
        )
        filter2 = Filter(
            pubkey_refs=[self.pk1.public_key.hex()],
        )
        filters = Filters([filter1, filter2])

        # Should match the entire pk1 thread and the DM exchange
        for event in self.pk1_thread + self.pk1_pk2_dms:
            assert filters.match(event)

        # Should not match anything in pk2's solo thread
        assert filters.match(self.pk2_thread[0]) is False
        assert filters.match(self.pk2_thread[1]) is False



import json
import unittest
import uuid

from nostr.event import Event
from nostr.message_pool import MessagePool


class TestMessagePool(unittest.TestCase):
    def test_empty_pool(self):
        mp = MessagePool()
        self.assertEqual(mp.has_events(), 0)
        self.assertEqual(mp.has_notices(), 0)
        self.assertEqual(mp.has_eose_notices(), 0)
        self.assertEqual(mp.has_ok_notices(), 0)

    def test_notices(self):
        mp = MessagePool()
        url = "ws://relay"
        mp.add_message('["NOTICE", "Test Notice"]', url)
        self.assertEqual(mp.has_events(), 0)
        self.assertEqual(mp.has_notices(), 1)
        self.assertEqual(mp.has_eose_notices(), 0)
        self.assertEqual(mp.has_ok_notices(), 0)
        results = mp.get_all()["notices"]
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].url, url)
        self.assertEqual(results[0].content, "Test Notice")

    def test_eose(self):
        mp = MessagePool()
        url = "ws://relay"
        mp.add_message(json.dumps(["EOSE", uuid.uuid1().hex]), url)
        self.assertEqual(mp.has_events(), 0)
        self.assertEqual(mp.has_notices(), 0)
        self.assertEqual(mp.has_ok_notices(), 0)
        self.assertEqual(mp.has_eose_notices(), 1)
        results = mp.get_all()["eose"]
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].url, url)

    def test_event(self):
        mp = MessagePool()
        e = Event()
        url = "ws://relay"
        mp.add_message(json.dumps(["EVENT", uuid.uuid1().hex, e.to_dict()]), url)
        self.assertEqual(mp.has_events(), 1)
        self.assertEqual(mp.has_notices(), 0)
        self.assertEqual(mp.has_eose_notices(), 0)
        self.assertEqual(mp.has_ok_notices(), 0)
        results = mp.get_all()["events"]
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].url, url)

    def test_ok(self):
        mp = MessagePool()
        url = "ws://relay"
        mp.add_message('["OK", "Test OK"]', url)
        self.assertEqual(mp.has_events(), 0)
        self.assertEqual(mp.has_notices(), 0)
        self.assertEqual(mp.has_eose_notices(), 0)
        self.assertEqual(mp.has_ok_notices(), 1)
        results = mp.get_all()["ok"]
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].url, url)
        self.assertEqual(results[0].content, '["OK", "Test OK"]')



import click
from click_aliases import ClickAliasedGroup

from nostr.key import PrivateKey, PublicKey


@click.group(cls=ClickAliasedGroup)
def cli():
    """Command related to key(s)."""


@cli.command(aliases=['gen', 'new'])
def create():
    """Creates a private and public key."""
    private_key = PrivateKey()
    public_key = private_key.public_key
    click.echo(f"Private key: {private_key.bech32()}")
    click.echo(f"Public key: {public_key.bech32()}")


@cli.command()
@click.option("-i", "--identifier", required=True, type=str)
def convert(identifier: str):
    """Converts npub key to hex."""
    if "npub" in identifier:
        public_key = PublicKey.from_npub(identifier)
    else:
        public_key = PublicKey.from_hex(identifier)
    click.echo(f"npub: {public_key.bech32()}")
    click.echo(f"hex: {public_key.hex()}")



import logging
import os

import click

from nostr._version import __version__

state = {"verbose": 3}

plugin_folder = os.path.join(os.path.dirname(__file__), "commands")


class CLI(click.MultiCommand):
    def list_commands(self, ctx):
        rv = []
        for filename in os.listdir(plugin_folder):
            if filename.endswith(".py") and filename != "__init__.py":
                rv.append(filename[:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        ns = {}
        fn = os.path.join(plugin_folder, name + ".py")
        if not os.path.exists(fn):
            return
        with open(fn) as f:
            code = compile(f.read(), fn, "exec")
            eval(code, ns, ns)
        return ns["cli"]


@click.version_option(__version__, "-v", "--version")
@click.command(cls=CLI)
def cli(verbose: int = 3):
    """CLI for nostr."""
    # Logging
    state["verbose"] = verbose
    log = logging.getLogger(__name__)
    verbosity = ["critical", "error", "warn", "info", "debug"][int(min(verbose, 4))]
    log.setLevel(getattr(logging, verbosity.upper()))
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    ch = logging.StreamHandler()
    ch.setLevel(getattr(logging, verbosity.upper()))
    ch.setFormatter(formatter)
    log.addHandler(ch)


if __name__ == "__main__":
    cli()



import json
from dataclasses import dataclass
from queue import Queue
from threading import Lock
from typing import List, Optional

from .event import Event
from .message_type import RelayMessageType


class EventMessage:
    def __init__(self, event: Event, subscription_id: str, url: str) -> None:
        self.event = event
        self.subscription_id = subscription_id
        self.url = url

    def __repr__(self):
        return f'EventMessage({self.url}: kind {str(self.event.kind)})'


class NoticeMessage:
    def __init__(self, content: str, url: str) -> None:
        self.content = content
        self.url = url

    def __repr__(self):
        return f'Notice({self.url}: {self.content})'


class EndOfStoredEventsMessage:
    def __init__(self, subscription_id: str, url: str) -> None:
        self.subscription_id = subscription_id
        self.url = url

    def __repr__(self):
        return f'EOSE({self.url})'


class OkMessage:
    def __init__(self, content: str, url: str) -> None:
        self.content = content
        self.url = url


class MessagePool:
    def __init__(self) -> None:
        self.events: Queue[EventMessage] = Queue()
        self.notices: Queue[NoticeMessage] = Queue()
        self.eose_notices: Queue[EndOfStoredEventsMessage] = Queue()
        self.ok_notices: Queue[OkMessage] = Queue()
        self._unique_events: set = set()
        self.lock: Lock = Lock()

    def add_message(self, message: str, url: str):
        self._process_message(message, url)

    def get_all(self):
        results = {"events": [], "notices": [], "eose": [], "ok": []}
        while self.has_events():
            results["events"].append(self.get_event())
        while self.has_notices():
            results["notices"].append(self.get_notice())
        while self.has_eose_notices():
            results["eose"].append(self.get_eose_notice())
        while self.has_ok_notices():
            results["ok"].append(self.get_ok_notice())
        return results

    def get_event(self):
        return self.events.get()

    def get_notice(self):
        return self.notices.get()

    def get_eose_notice(self):
        return self.eose_notices.get()

    def get_ok_notice(self):
        return self.ok_notices.get()

    def has_events(self):
        return self.events.qsize() > 0

    def has_notices(self):
        return self.notices.qsize() > 0

    def has_eose_notices(self):
        return self.eose_notices.qsize() > 0

    def has_ok_notices(self):
        return self.ok_notices.qsize() > 0

    def _process_message(self, message: str, url: str):
        message_json = json.loads(message)
        message_type = message_json[0]
        if message_type == RelayMessageType.EVENT:
            subscription_id = message_json[1]
            event = Event.from_dict(message_json[2])
            with self.lock:
                uid = subscription_id + event.id
                if uid not in self._unique_events:
                    self.events.put(EventMessage(event, subscription_id, url))
                    self._unique_events.add(uid)
        elif message_type == RelayMessageType.NOTICE:
            self.notices.put(NoticeMessage(message_json[1], url))
        elif message_type == RelayMessageType.END_OF_STORED_EVENTS:
            self.eose_notices.put(EndOfStoredEventsMessage(message_json[1], url))
        elif message_type == RelayMessageType.OK:
            self.ok_notices.put(OkMessage(message, url))

    def __repr__(self):
        return (
            f'Pool(events({self.events.qsize()}) '
            f'notices({self.notices.qsize()}) '
            f'eose({self.eose_notices.qsize()})) '
            f'ok({self.ok_notices.qsize()}))'
        )


@dataclass
class EventMessageStore:
    eventMessages: Optional[List[EventMessage]] = None

    def add_event(self, event):
        if self.eventMessages is None:
            self.eventMessages = []
        if isinstance(event, list):
            self.eventMessages += event
        else:
            self.eventMessages.append(event)

    def get_newest_event(self):
        if not self.eventMessages:
            return None
        return max(self.eventMessages, key=lambda x: x.event.date_time())

    def get_events_by_url(self, url):
        return [event for event in self.eventMessages if event.url == url]

    def get_events_by_id(self, subscription_id):
        return [
            event
            for event in self.eventMessages
            if event.subscription_id == subscription_id
        ]

    def __repr__(self):
        if not self.eventMessages:
            return 'EventMessageStore()'
        return f'EventMessageStore({len(self.eventMessages)} events)'



import unittest

from nostr import pow
from nostr.event import Event, EventKind
from nostr.key import PrivateKey
from nostr.pow import Pow, PowEvent, PowKey, PowVanityKey


class TestPow(unittest.TestCase):
    def test_count_leading_zero(self):
        self.assertEqual(pow.count_leading_zero_bits("048d"), 5)
        self.assertEqual(pow.count_leading_zero_bits("ffff"), 0)

    def test_mine_event(self):
        """Test mining an event with specific difficulty."""
        public_key = PrivateKey().public_key.hex()
        difficulty = 8
        event = Event(content='test', public_key=public_key, kind=EventKind.TEXT_NOTE)
        p = PowEvent(difficulty, event)
        event = p.mine()
        self.assertTrue(pow.count_leading_zero_bits(event.id) >= difficulty)

    def test_mine_key(self):
        """Test mining a public key with specific difficulty."""
        difficulty = 8
        p = PowKey(difficulty)
        sk = p.mine()
        self.assertTrue(pow.count_leading_zero_bits(sk.public_key.hex()) >= difficulty)

    def test_time_estimates(self):
        """Test functions to estimate POW time."""
        public_key = PrivateKey().public_key.hex()
        event = Event(content='test', public_key=public_key, kind=EventKind.TEXT_NOTE)
        p = PowEvent(8, event)
        # test successful run of all estimators
        p.get_expected_time()
        p_key = PowKey(8)
        p_key.get_expected_time()
        p_vkey = PowVanityKey("t")
        p_vkey.get_expected_time()

    def test_mine_vanity_key(self):
        """Test vanity key mining."""
        pattern = '23'
        p = PowVanityKey(pattern)
        sk = p.mine()
        sk.public_key.bech32()
        self.assertTrue(sk.public_key.bech32().startswith(f'npub1{pattern}'))

        p = PowVanityKey(suffix=pattern)
        sk = p.mine()
        self.assertTrue(sk.public_key.bech32().endswith(pattern))

        # mine an invalid pattern
        pattern = '1'

        with self.assertRaisesRegex(ValueError, "not in valid list of bech32 chars"):
            p = PowVanityKey(pattern)

    def test_expected_pow_guesses(self):
        p = Pow()
        p.n_pattern = 32
        p.n_options = 2
        e1 = p.get_expected_guesses()

        p.n_pattern = 8
        p.n_options = 16
        e2 = p.get_expected_guesses()
        self.assertEqual(e1, e2)
